@@ -1,3 +1,5 @@
+/* vim: set ts=8 sw=8 noexpandtab: */
+
 /* The CSS Theme Engine for Gtk+.
  * Copyright (C) 2008 Robert Staudinger
  *
@@ -20,54 +22,15 @@
 #include <gtk/gtk.h>
 #include "gce-maps.h"
 #include "gce-node.h"
-#include "gce-rc-style.h"
 #include "gce-style.h"
-
-struct GceStyle_
-{
-	GtkStyle parent;
-};
-
-struct GceStyleClass_
-{
-	GtkStyleClass parent;
-};
+#include "config.h"
 
 static GType          gce_style_type = 0;
 static GtkStyleClass *gce_style_parent_class = NULL;
 
-static ccss_selector_group_t *
-setup (GceNode const *node, 
-       GceNode const *base)
-{
-	ccss_stylesheet_t const	*stylesheet;
-	ccss_selector_group_t	*group;
-	gboolean		 status;
-
-	stylesheet = gce_rc_style_get_stylesheet ();
-	group = ccss_selector_group_new ();
-	status = FALSE;
-
-	/* query for widget theming */
-	status |= ccss_stylesheet_query_collect (stylesheet, (ccss_node_t const *) node, group, FALSE);
-
-	/* query for primitive theming */
-	status |= ccss_stylesheet_query_collect (stylesheet, (ccss_node_t const *) base, group, TRUE);
-
-	if (!status) {
-		g_warning ("Un-themed widget `%s', primitive `%s'.", 
-			   G_OBJECT_TYPE_NAME (G_OBJECT (gce_node_get_widget (node))), gce_node_get_primitive (base));
-		ccss_selector_group_free (group);
-		group = NULL;
-	}
-
-	return group;
-}
-
 static void
-rectangle (GtkStyle		*self, 
-	   GceNode const	*node,
-	   GceNode const	*base,
+rectangle (GceStyle		*self, 
+	   GceNode 		*node,
 	   GdkWindow		*window, 
 	   GdkRectangle		*area, 
 	   GtkWidget		*widget, 
@@ -77,14 +40,20 @@ rectangle (GtkStyle		*self,
 	   gint			 height,
 	   gboolean		 fill)
 {
-	ccss_selector_group_t	*group;
-	ccss_style_t		 style;
-	cairo_t			*cr;
+	cairo_t		*cr;
+	ccss_style_t	*style = NULL;
 
-	group = setup (node, base);
-	if (group) {
+	g_return_if_fail (self);
+	g_return_if_fail (self->stylesheet);
+
+	if (node) {
+		style = ccss_stylesheet_query (self->stylesheet, (ccss_node_t *) node);
+	} else if (self->selector) {
+		style = ccss_stylesheet_query_type (self->stylesheet, self->selector);
+	}
+
+	if (style) {
 		cr = gdk_cairo_create (window);
-		ccss_style_init (&style);
 
 		if (area) {
 			gdk_cairo_rectangle (cr, area);
@@ -99,28 +68,20 @@ rectangle (GtkStyle		*self,
 			height = height == -1 ? h : height;
 		}
 
-		ccss_selector_group_apply (group, &style);
-
-		/* Hackishly support "background-attachment: fixed". */
-		ccss_style_set_viewport (&style,
-			widget->allocation.x, widget->allocation.y,
-			widget->allocation.width, widget->allocation.height);
-
 		if (fill) {
-			ccss_style_draw_rectangle (&style, cr, x, y, width, height);
+			ccss_cairo_style_draw_rectangle (style, cr, x, y, width, height);
 		} else {
-			ccss_style_draw_outline (&style, cr, x, y, width, height);
+			ccss_cairo_style_draw_outline (style, cr, x, y, width, height);
 		}
 
-		cairo_destroy (cr);
-		ccss_selector_group_free (group);
+		cairo_destroy (cr), cr = NULL;
+		ccss_style_destroy (style), style = NULL;
 	}
 }
 
 static void
-gap (GtkStyle		*self, 
-     GceNode const	*node,
-     GceNode const	*base,
+gap (GceStyle		*self,
+     GceNode 		*node,
      GdkWindow		*window, 
      GdkRectangle	*area, 
      GtkWidget		*widget, 
@@ -132,14 +93,20 @@ gap (GtkStyle		*self,
      gint		 gap_start,
      gint		 gap_width)
 {
-	ccss_selector_group_t	*group;
-	ccss_style_t		 style;
-	cairo_t			*cr;
+	cairo_t		*cr;
+	ccss_style_t	*style = NULL;
 
-	group = setup (node, base);
-	if (group) {
+	g_return_if_fail (self);
+	g_return_if_fail (self->stylesheet);
+
+	if (node) {
+		style = ccss_stylesheet_query (self->stylesheet, (ccss_node_t *) node);
+	} else if (self->selector) {
+		style = ccss_stylesheet_query_type (self->stylesheet, self->selector);
+	}
+
+	if (style) {
 		cr = gdk_cairo_create (window);
-		ccss_style_init (&style);
 
 		if (area) {
 			gdk_cairo_rectangle (cr, area);
@@ -154,24 +121,18 @@ gap (GtkStyle		*self,
 			height = height == -1 ? h : height;
 		}
 
-		ccss_selector_group_apply (group, &style);
+		ccss_cairo_style_draw_rectangle_with_gap (style, cr, x, y,
+							  width, height,
+							  gap_side, gap_start, gap_width);
 
-		/* Hackishly support "background-attachment: fixed". */
-		ccss_style_set_viewport (&style,
-			widget->allocation.x, widget->allocation.y,
-			widget->allocation.width, widget->allocation.height);
-
-		ccss_style_draw_gap (&style, cr, x, y, width, height, gap_side, gap_start, gap_width);
-
-		cairo_destroy (cr);
-		ccss_selector_group_free (group);
+		cairo_destroy (cr), cr = NULL;
+		ccss_style_destroy (style), style = NULL;
 	}
 }
 
 static void
-line (GtkStyle		*self, 
+line (GceStyle		*self,
       GceNode		*node,
-      GceNode		*base,
       GdkWindow		*window, 
       GdkRectangle	*area, 
       GtkWidget		*widget, 
@@ -180,31 +141,30 @@ line (GtkStyle		*self,
       gint		 y1, 
       gint		 y2)
 {
-	ccss_selector_group_t	*group;
-	ccss_style_t		 style;
-	cairo_t			*cr;
+	cairo_t		*cr;
+	ccss_style_t	*style = NULL;
 
-	group = setup (node, base);
-	if (group) {
+	g_return_if_fail (self);
+	g_return_if_fail (self->stylesheet);
+
+	if (node) {
+		style = ccss_stylesheet_query (self->stylesheet, (ccss_node_t *) node);
+	} else if (self->selector) {
+		style = ccss_stylesheet_query_type (self->stylesheet, self->selector);
+	}
+
+	if (style) {
 		cr = gdk_cairo_create (window);
-		ccss_style_init (&style);
 
 		if (area) {
 			gdk_cairo_rectangle (cr, area);
 			cairo_clip (cr);
 		}
 
-		ccss_selector_group_apply (group, &style);
+		ccss_cairo_style_draw_line (style, cr, x1, x2, y1, y2);
 
-		/* Hackishly support "background-attachment: fixed". */
-		ccss_style_set_viewport (&style,
-			widget->allocation.x, widget->allocation.y,
-			widget->allocation.width, widget->allocation.height);
-
-		ccss_style_draw_line (&style, cr, x1, x2, y1, y2);
-
-		cairo_destroy (cr);
-		ccss_selector_group_free (group);
+		cairo_destroy (cr), cr = NULL;
+		ccss_style_destroy (style), style = NULL;
 	}
 }
 
@@ -219,24 +179,22 @@ draw_hline (GtkStyle		*self,
 	    gint		 x2, 
 	    gint		 y)
 {
-	GceNode *node;
-	GceNode *base;
+	char const	*role;
+	GceNode		*node = NULL;
 
-	base = gce_node_cache_fetch_primitive_node (widget, 
-			gce_maps_get_detail (detail),
+	detail = gce_maps_get_detail (detail);
+	role = gce_maps_match_role (widget, "hline", -1, detail, x1, y, x2 - x1, -1);
+
+	if (widget)
+		node = gce_node_cache_fetch_node (widget, 
+			detail,
 			gce_maps_get_state (state),
 			NULL, NULL, NULL, NULL, NULL,
-			"hline");
+			role ? role : "hline");
 
-	node = gce_node_cache_fetch_node (widget,
-			gce_maps_get_detail (detail),
-			gce_maps_get_state (state),
-			NULL, NULL, NULL, NULL, NULL);
+	line (GCE_STYLE (self), node, window, area, widget, x1, x2, y, y);
 
-	line (self, node, base, window, area, widget, x1, x2, y, y);
-
-	gce_node_cache_release_node (node);
-	gce_node_cache_release_node (base);
+	gce_node_cache_release_node (node), node = NULL;
 }
 
 static void 
@@ -250,24 +208,22 @@ draw_vline (GtkStyle		*self,
 	    gint		 y2, 
 	    gint		 x)
 {
-	GceNode *node;
-	GceNode *base;
+	char const	*role;
+	GceNode		*node = NULL;
 
-	base = gce_node_cache_fetch_primitive_node (widget, 
-			gce_maps_get_detail (detail),
+	detail = gce_maps_get_detail (detail);
+	role = gce_maps_match_role (widget, "vline", -1, detail, x, y1, -1, y2 - y1);
+
+	if (widget)
+		node = gce_node_cache_fetch_node (widget, 
+			detail,
 			gce_maps_get_state (state),
 			NULL, NULL, NULL, NULL, NULL,
-			"vline");
+			role ? role : "vline");
 
-	node = gce_node_cache_fetch_node (widget,
-			gce_maps_get_detail (detail),
-			gce_maps_get_state (state),
-			NULL, NULL, NULL, NULL, NULL);
+	line (GCE_STYLE (self), node, window, area, widget, x, x, y1, y2);
 
-	line (self, node, base, window, area, widget, x, x, y1, y2);
-
-	gce_node_cache_release_node (node);
-	gce_node_cache_release_node (base);
+	gce_node_cache_release_node (node), node = NULL;
 
 }
 
@@ -284,27 +240,24 @@ draw_shadow (GtkStyle		*self,
 	     gint		 width, 
 	     gint		 height)
 {
-	GceNode *node;
-	GceNode *base;
+	char const	*role;
+	GceNode		*node = NULL;
 
-	base = gce_node_cache_fetch_primitive_node (widget, 
-			gce_maps_get_detail (detail),
+	detail = gce_maps_get_detail (detail);
+	role = gce_maps_match_role (widget, "shadow", shadow, detail, x, y, width, height);
+
+	if (widget)
+		node = gce_node_cache_fetch_node (widget, 
+			detail,
 			gce_maps_get_state (state),
 			gce_maps_get_shadow (shadow),
 			NULL, NULL, NULL, NULL,
-			"shadow");
+			role ? role : "shadow");
 
-	node = gce_node_cache_fetch_node (widget, 
-			gce_maps_get_detail (detail),
-			gce_maps_get_state (state),
-			gce_maps_get_shadow (shadow),
-			NULL, NULL, NULL, NULL);
-
-	rectangle (self, node, base, window, area, widget, 
+	rectangle (GCE_STYLE (self), node, window, area, widget, 
 		   x, y, width, height, TRUE);
 
-	gce_node_cache_release_node (node);
-	gce_node_cache_release_node (base);
+	gce_node_cache_release_node (node), node = NULL;
 }
 
 static void 
@@ -320,27 +273,24 @@ draw_box (GtkStyle		*self,
 	  gint			 width,
 	  gint			 height)
 {
-	GceNode *node;
-	GceNode *base;
+	char const	*role;
+	GceNode		*node = NULL;
 
-	base = gce_node_cache_fetch_primitive_node (widget, 
-			gce_maps_get_detail (detail),
+	detail = gce_maps_get_detail (detail);
+	role = gce_maps_match_role (widget, "box", shadow, detail, x, y, width, height);
+
+	if (widget)
+		node = gce_node_cache_fetch_node (widget, 
+			detail,
 			gce_maps_get_state (state),
 			gce_maps_get_shadow (shadow),
 			NULL, NULL, NULL, NULL,
-			"box");
+			role ? role : "box");
 
-	node = gce_node_cache_fetch_node (widget, 
-			gce_maps_get_detail (detail),
-			gce_maps_get_state (state),
-			gce_maps_get_shadow (shadow),
-			NULL, NULL, NULL, NULL);
-
-	rectangle (self, node, base, window, area, widget,
+	rectangle (GCE_STYLE (self), node, window, area, widget,
 		   x, y, width, height, TRUE);
 
-	gce_node_cache_release_node (node);
-	gce_node_cache_release_node (base);
+	gce_node_cache_release_node (node), node = NULL;
 }
 
 static void 
@@ -356,27 +306,24 @@ draw_flat_box (GtkStyle		*self,
 	       gint		 width,
 	       gint		 height)
 {
-	GceNode *node;
-	GceNode *base;
+	char const	*role;
+	GceNode		*node = NULL;
 
-	base = gce_node_cache_fetch_primitive_node (widget, 
-			gce_maps_get_detail (detail),
+	detail = gce_maps_get_detail (detail);
+	role = gce_maps_match_role (widget, "flat-box", shadow, detail, x, y, width, height);
+
+	if (widget)
+		node = gce_node_cache_fetch_node (widget, 
+			detail,
 			gce_maps_get_state (state),
 			gce_maps_get_shadow (shadow),
 			NULL, NULL, NULL, NULL,
-			"flatbox");
+			role ? role : "flat-box");
 
-	node = gce_node_cache_fetch_node (widget, 
-			gce_maps_get_detail (detail),
-			gce_maps_get_state (state),
-			gce_maps_get_shadow (shadow),
-			NULL, NULL, NULL, NULL);
-
-	rectangle (self, node, base, window, area, widget, 
+	rectangle (GCE_STYLE (self), node, window, area, widget, 
 		   x, y, width, height, TRUE);
 
-	gce_node_cache_release_node (node);
-	gce_node_cache_release_node (base);
+	gce_node_cache_release_node (node), node = NULL;
 }
 
 static void 
@@ -392,27 +339,24 @@ draw_check (GtkStyle		*self,
 	    gint		 width,
 	    gint		 height)
 {
-	GceNode *node;
-	GceNode *base;
+	char const	*role;
+	GceNode		*node = NULL;
 
-	base = gce_node_cache_fetch_primitive_node (widget, 
-			gce_maps_get_detail (detail),
+	detail = gce_maps_get_detail (detail);
+	role = gce_maps_match_role (widget, "check", shadow, detail, x, y, width, height);
+
+	if (widget)
+		node = gce_node_cache_fetch_node (widget, 
+			detail,
 			gce_maps_get_state (state),
 			gce_maps_get_shadow (shadow),
 			NULL, NULL, NULL, NULL,
-			"check");
+			role ? role : "check");
 
-	node = gce_node_cache_fetch_node (widget, 
-			gce_maps_get_detail (detail),
-			gce_maps_get_state (state),
-			gce_maps_get_shadow (shadow),
-			NULL, NULL, NULL, NULL);
-
-	rectangle (self, node, base, window, area, widget, 
+	rectangle (GCE_STYLE (self), node, window, area, widget, 
 		   x, y, width, height, TRUE);
 
-	gce_node_cache_release_node (node);
-	gce_node_cache_release_node (base);
+	gce_node_cache_release_node (node), node = NULL;
 }
 
 static void 
@@ -428,27 +372,24 @@ draw_option (GtkStyle		 *self,
 	     gint		 width,
 	     gint		 height)
 {
-	GceNode *node;
-	GceNode *base;
+	char const	*role;
+	GceNode		*node = NULL;
 
-	base = gce_node_cache_fetch_primitive_node (widget, 
-			gce_maps_get_detail (detail),
+	detail = gce_maps_get_detail (detail);
+	role = gce_maps_match_role (widget, "option", shadow, detail, x, y, width, height);
+
+	if (widget)
+		node = gce_node_cache_fetch_node (widget, 
+			detail,
 			gce_maps_get_state (state),
 			gce_maps_get_shadow (shadow),
 			NULL, NULL, NULL, NULL,
-			"option");
+			role ? role : "option");
 
-	node = gce_node_cache_fetch_node (widget, 
-			gce_maps_get_detail (detail),
-			gce_maps_get_state (state),
-			gce_maps_get_shadow (shadow),
-			NULL, NULL, NULL, NULL);
-
-	rectangle (self, node, base, window, area, widget, 
+	rectangle (GCE_STYLE (self), node, window, area, widget, 
 		   x, y, width, height, TRUE);
 
-	gce_node_cache_release_node (node);
-	gce_node_cache_release_node (base);
+	gce_node_cache_release_node (node), node = NULL;
 }
 
 static void 
@@ -467,27 +408,24 @@ draw_shadow_gap (GtkStyle		*self,
 		 gint			 gap_start,
 		 gint			 gap_width)
 {
-	GceNode *node;
-	GceNode *base;
+	char const	*role;
+	GceNode		*node = NULL;
 
-	base = gce_node_cache_fetch_primitive_node (widget, 
-			gce_maps_get_detail (detail),
+	detail = gce_maps_get_detail (detail);
+	role = gce_maps_match_role (widget, "shadow-gap", shadow, detail, x, y, width, height);
+
+	if (widget)
+		node = gce_node_cache_fetch_node (widget, 
+			detail,
 			gce_maps_get_state (state),
 			gce_maps_get_shadow (shadow),
 			NULL, gce_maps_get_position (gap_side), NULL, NULL,
-			"shadow");
+			role ? role : "shadow");
 
-	node = gce_node_cache_fetch_node (widget, 
-			gce_maps_get_detail (detail),
-			gce_maps_get_state (state),
-			gce_maps_get_shadow (shadow),
-			NULL, gce_maps_get_position (gap_side), NULL, NULL);
-
-	gap (self, node, base, window, area, widget, 
+	gap (GCE_STYLE (self), node, window, area, widget, 
 	     x, y, width, height, gap_side, gap_start, gap_width);
 
-	gce_node_cache_release_node (node);
-	gce_node_cache_release_node (base);
+	gce_node_cache_release_node (node), node = NULL;
 }
 
 static void 
@@ -506,27 +444,24 @@ draw_box_gap (GtkStyle			*self,
 	      gint			 gap_start,
 	      gint			 gap_width)
 {
-	GceNode *node;
-	GceNode *base;
+	char const	*role;
+	GceNode		*node = NULL;
 
-	base = gce_node_cache_fetch_primitive_node (widget, 
-			gce_maps_get_detail (detail),
+	detail = gce_maps_get_detail (detail);
+	role = gce_maps_match_role (widget, "box-gap", shadow, detail, x, y, width, height);
+
+	if (widget)
+		node = gce_node_cache_fetch_node (widget, 
+			detail,
 			gce_maps_get_state (state),
 			gce_maps_get_shadow (shadow),
 			NULL, gce_maps_get_position (gap_side), NULL, NULL,
-			"box");
+			role ? role : "box");
 
-	node = gce_node_cache_fetch_node (widget, 
-			gce_maps_get_detail (detail),
-			gce_maps_get_state (state),
-			gce_maps_get_shadow (shadow),
-			NULL, gce_maps_get_position (gap_side), NULL, NULL);
-
-	gap (self, node, base, window, area, widget, 
+	gap (GCE_STYLE (self), node, window, area, widget, 
 	     x, y, width, height, gap_side, gap_start, gap_width);
 
-	gce_node_cache_release_node (node);
-	gce_node_cache_release_node (base);
+	gce_node_cache_release_node (node), node = NULL;
 }
 
 static void 
@@ -543,22 +478,20 @@ draw_extension (GtkStyle	*self,
 		gint		 height,
 		GtkPositionType	 gap_side)
 {
-	GceNode *node;
-	GceNode *base;
-	gint	 gap_width;
+	char const	*role;
+	gint		 gap_width;
+	GceNode		*node = NULL;
 
-	base = gce_node_cache_fetch_primitive_node (widget, 
-			gce_maps_get_detail (detail),
+	detail = gce_maps_get_detail (detail);
+	role = gce_maps_match_role (widget, "extension", shadow, detail, x, y, width, height);
+
+	if (widget)
+		node = gce_node_cache_fetch_node (widget, 
+			detail,
 			gce_maps_get_state (state),
 			gce_maps_get_shadow (shadow),
 			NULL, gce_maps_get_position (gap_side), NULL, NULL,
-			"extension");
-
-	node = gce_node_cache_fetch_node (widget, 
-			gce_maps_get_detail (detail),
-			gce_maps_get_state (state),
-			gce_maps_get_shadow (shadow),
-			NULL, gce_maps_get_position (gap_side), NULL, NULL);
+			role ? role : "extension");
 
 	switch (gap_side) {
 	case GTK_POS_LEFT:
@@ -571,11 +504,10 @@ draw_extension (GtkStyle	*self,
 		break;
 	}
 
-	gap (self, node, base, window, area, widget, 
+	gap (GCE_STYLE (self), node, window, area, widget, 
 	     x, y, width, height, gap_side, 0, gap_width);
 
-	gce_node_cache_release_node (node);
-	gce_node_cache_release_node (base);
+	gce_node_cache_release_node (node), node = NULL;
 }
 
 static void 
@@ -592,29 +524,25 @@ draw_slider (GtkStyle		*self,
 	     gint		 height,
 	     GtkOrientation	 orientation)
 {
-	GceNode *node;
-	GceNode *base;
+	char const	*role;
+	GceNode		*node = NULL;
 
-	base = gce_node_cache_fetch_primitive_node (widget,
-			gce_maps_get_detail (detail),
+	detail = gce_maps_get_detail (detail);
+	role = gce_maps_match_role (widget, "slider", shadow, detail, x, y, width, height);
+
+	if (widget)
+		node = gce_node_cache_fetch_node (widget,
+			detail,
 			gce_maps_get_state (state),
 			gce_maps_get_shadow (shadow),
 			gce_maps_get_orientation (orientation),
 			NULL, NULL, NULL,
-			"slider");
+			role ? role : "slider");
 
-	node = gce_node_cache_fetch_node (widget, 
-			gce_maps_get_detail (detail),
-			gce_maps_get_state (state),
-			gce_maps_get_shadow (shadow),
-			gce_maps_get_orientation (orientation),
-			NULL, NULL, NULL);
-
-	rectangle (self, node, base, window, area, widget, 
+	rectangle (GCE_STYLE (self), node, window, area, widget, 
 		   x, y, width, height, TRUE);
 
-	gce_node_cache_release_node (node);
-	gce_node_cache_release_node (base);
+	gce_node_cache_release_node (node), node = NULL;
 }
 
 static void 
@@ -631,29 +559,25 @@ draw_handle (GtkStyle		*self,
 	     gint		 height,
 	     GtkOrientation	 orientation)
 {
-	GceNode *node;
-	GceNode *base;
+	char const	*role;
+	GceNode		*node = NULL;
 
-	base = gce_node_cache_fetch_primitive_node (widget,
-			gce_maps_get_detail (detail),
+	detail = gce_maps_get_detail (detail);
+	role = gce_maps_match_role (widget, "handle", shadow, detail, x, y, width, height);
+
+	if (widget)
+		node = gce_node_cache_fetch_node (widget,
+			detail,
 			gce_maps_get_state (state),
 			gce_maps_get_shadow (shadow),
 			gce_maps_get_orientation (orientation),
 			NULL, NULL, NULL,
-			"handle");
+			role ? role : "handle");
 
-	node = gce_node_cache_fetch_node (widget, 
-			gce_maps_get_detail (detail),
-			gce_maps_get_state (state),
-			gce_maps_get_shadow (shadow),
-			gce_maps_get_orientation (orientation),
-			NULL, NULL, NULL);
-
-	rectangle (self, node, base, window, area, widget, 
+	rectangle (GCE_STYLE (self), node, window, area, widget, 
 		   x, y, width, height, TRUE);
 
-	gce_node_cache_release_node (node);
-	gce_node_cache_release_node (base);
+	gce_node_cache_release_node (node), node = NULL;
 }
 
 static void 
@@ -669,25 +593,23 @@ draw_resize_grip (GtkStyle		*self,
 		  gint			 width,
 		  gint			 height)
 {
-	GceNode *node;
-	GceNode *base;
+	char const	*role;
+	GceNode		*node = NULL;
 
-	base = gce_node_cache_fetch_primitive_node (widget,
-			gce_maps_get_detail (detail),
+	detail = gce_maps_get_detail (detail);
+	role = gce_maps_match_role (widget, "resize-grip", -1, detail, x, y, width, height);
+
+	if (widget)
+		node = gce_node_cache_fetch_node (widget,
+			detail,
 			gce_maps_get_state (state),
 			NULL, NULL, NULL, gce_maps_get_window_edge (edge), NULL,
-			"resizegrip");
+			role ? role : "resizegrip");
 
-	node = gce_node_cache_fetch_node (widget, 
-			gce_maps_get_detail (detail),
-			gce_maps_get_state (state),
-			NULL, NULL, NULL, gce_maps_get_window_edge (edge), NULL);
-
-	rectangle (self, node, base, window, area, widget, 
+	rectangle (GCE_STYLE (self), node, window, area, widget, 
 		   x, y, width, height, TRUE);
 
-	gce_node_cache_release_node (node);
-	gce_node_cache_release_node (base);
+	gce_node_cache_release_node (node), node = NULL;
 }
 
 static void 
@@ -702,27 +624,26 @@ draw_focus (GtkStyle		*self,
 	    gint		 width,
 	    gint		 height)
 {
-	GceNode *node;
-	GceNode *base;
+	char const	*role;
+	GceNode		*node = NULL;
 
-	base = gce_node_cache_fetch_primitive_node (widget,
-			gce_maps_get_detail (detail),
+	detail = gce_maps_get_detail (detail);
+	role = gce_maps_match_role (widget, "focus", -1, detail, x, y, width, height);
+
+	if (widget)
+		node = gce_node_cache_fetch_node (widget,
+			detail,
 			gce_maps_get_state (state),
 			NULL, NULL, NULL, NULL, NULL,
-			"focus");
+			role ? role : "focus");
 
-	node = gce_node_cache_fetch_node (widget, 
-			gce_maps_get_detail (detail),
-			gce_maps_get_state (state),
-			NULL, NULL, NULL, NULL, NULL);
-
-	rectangle (self, node, base, window, area, widget, 
+	rectangle (GCE_STYLE (self), node, window, area, widget, 
 		   x, y, width, height, FALSE);
 
-	gce_node_cache_release_node (node);
-	gce_node_cache_release_node (base);
+	gce_node_cache_release_node (node), node = NULL;
 }
 
+#if 0
 static void
 draw_expander (GtkStyle		*self,
 	       GdkWindow	*window,
@@ -735,9 +656,12 @@ draw_expander (GtkStyle		*self,
 	       GtkExpanderStyle	 style)
 {
 #define DEFAULT_EXPANDER_SIZE 12
-	GceNode *node;
-	GceNode *base;
-	gint	 expander_size;
+	char const	*role;
+	gint		 expander_size;
+	GceNode		*node = NULL;
+
+	detail = gce_maps_get_detail (detail);
+	role = gce_maps_match_role (widget, "expander", -1, detail, x, y, -1, -1);
 
 	if (widget &&
 	    gtk_widget_class_find_style_property (GTK_WIDGET_GET_CLASS (widget),
@@ -749,24 +673,20 @@ draw_expander (GtkStyle		*self,
 		expander_size = DEFAULT_EXPANDER_SIZE;
 	}
 
-	base = gce_node_cache_fetch_primitive_node (widget,
-			gce_maps_get_detail (detail),
+	if (widget)
+		node = gce_node_cache_fetch_node (widget,
+			detail,
 			gce_maps_get_state (state),
 			NULL, NULL, NULL, NULL, gce_maps_get_expander_style (style),
-			"expander");
+			role ? role : "expander");
 
-	node = gce_node_cache_fetch_node (widget, 
-			gce_maps_get_detail (detail),
-			gce_maps_get_state (state),
-			NULL, NULL, NULL, NULL, gce_maps_get_expander_style (style));
-
-	rectangle (self, node, base, window, area, widget, 
+	rectangle (GCE_STYLE (self), node, window, area, widget, 
 		   x, y, expander_size, expander_size, TRUE);
 
-	gce_node_cache_release_node (node);
-	gce_node_cache_release_node (base);
+	gce_node_cache_release_node (node), node = NULL;
 #undef DEFAULT_EXPANDER_SIZE
 }
+#endif
 
 static void
 draw_diamond (GtkStyle		*self,
@@ -781,27 +701,24 @@ draw_diamond (GtkStyle		*self,
 	      gint		 width,
 	      gint		 height)
 {
-	GceNode *node;
-	GceNode *base;
+	char const	*role;
+	GceNode		*node = NULL;
 
-	base = gce_node_cache_fetch_primitive_node (widget,
-			gce_maps_get_detail (detail),
+	detail = gce_maps_get_detail (detail);
+	role = gce_maps_match_role (widget, "diamond", shadow, detail, x, y, width, height);
+
+	if (widget)
+		node = gce_node_cache_fetch_node (widget,
+			detail,
 			gce_maps_get_state (state),
 			gce_maps_get_shadow (shadow),
 			NULL, NULL, NULL, NULL,
-			"diamond");
+			role ? role : "diamond");
 
-	node = gce_node_cache_fetch_node (widget, 
-			gce_maps_get_detail (detail),
-			gce_maps_get_state (state),
-			gce_maps_get_shadow (shadow),
-			NULL, NULL, NULL, NULL);
-
-	rectangle (self, node, base, window, area, widget, 
+	rectangle (GCE_STYLE (self), node, window, area, widget, 
 		   x, y, width, height, TRUE);
 
-	gce_node_cache_release_node (node);
-	gce_node_cache_release_node (base);
+	gce_node_cache_release_node (node), node = NULL;
 }
 
 /*
@@ -822,29 +739,25 @@ draw_arrow (GtkStyle		*self,
 	    gint		 width,
 	    gint		 height)
 {
-	GceNode *node;
-	GceNode *base;
+	char const	*role;
+	GceNode		*node = NULL;
 
-	base = gce_node_cache_fetch_primitive_node (widget,
-			gce_maps_get_detail (detail),
+	detail = gce_maps_get_detail (detail);
+	role = gce_maps_match_role (widget, "arrow", shadow, detail, x, y, width, height);
+
+	if (widget)
+		node = gce_node_cache_fetch_node (widget,
+			detail,
 			gce_maps_get_state (state),
 			gce_maps_get_shadow (shadow),
 			gce_maps_get_arrow (arrow),
 			NULL, NULL, NULL,
-			"arrow");
+			role ? role : "arrow");
 
-	node = gce_node_cache_fetch_node (widget, 
-			gce_maps_get_detail (detail),
-			gce_maps_get_state (state),
-			gce_maps_get_shadow (shadow),
-			gce_maps_get_arrow (arrow),
-			NULL, NULL, NULL);
-
-	rectangle (self, node, base, window, area, widget, 
+	rectangle (GCE_STYLE (self), node, window, area, widget, 
 		   x, y, width, height, TRUE);
 
-	gce_node_cache_release_node (node);
-	gce_node_cache_release_node (base);
+	gce_node_cache_release_node (node), node = NULL;
 }
 
 static void
@@ -860,42 +773,65 @@ draw_tab (GtkStyle		*self,
 	  gint			 width,
 	  gint			 height)
 {
-	GceNode *node;
-	GceNode *base;
+	char const	*role;
+	GceNode		*node = NULL;
 
-	base = gce_node_cache_fetch_primitive_node (widget,
-			gce_maps_get_detail (detail),
+	detail = gce_maps_get_detail (detail);
+	role = gce_maps_match_role (widget, "tab", shadow, detail, x, y, width, height);
+
+	if (widget)
+		node = gce_node_cache_fetch_node (widget,
+			detail,
 			gce_maps_get_state (state),
 			gce_maps_get_shadow (shadow),
 			NULL, NULL, NULL, NULL,
-			"tab");
+			role ? role : "tab");
 
-	node = gce_node_cache_fetch_node (widget, 
-			gce_maps_get_detail (detail),
-			gce_maps_get_state (state),
-			gce_maps_get_shadow (shadow),
-			NULL, NULL, NULL, NULL);
-
-	rectangle (self, node, base, window, area, widget, 
+	rectangle (GCE_STYLE (self), node, window, area, widget, 
 		   x, y, width, height, TRUE);
 
-	gce_node_cache_release_node (node);
-	gce_node_cache_release_node (base);
+	gce_node_cache_release_node (node), node = NULL;
+}
+
+static void
+finalize (GObject *instance)
+{
+	GceStyle *self;
+
+	self = GCE_STYLE (instance);
+
+	if (self->stylesheet) {
+		ccss_stylesheet_destroy (self->stylesheet);
+		self->stylesheet = NULL;
+	}
+
+	if (self->selector) {
+		g_free (self->selector);
+		self->selector = NULL;
+	}
+
+	G_OBJECT_CLASS (gce_style_parent_class)->finalize (instance);
 }
 
 static void 
 instance_init (GceStyle *self)
 {
-
+	self->stylesheet = NULL;
+	self->selector = NULL;
 }
 
 static void 
 class_init (GceStyleClass *klass)
 {
-	GtkStyleClass *style_class = GTK_STYLE_CLASS (klass);
+	GObjectClass	*go_class;
+	GtkStyleClass	*style_class;
 
 	gce_style_parent_class = g_type_class_peek_parent (klass);
 
+	go_class = G_OBJECT_CLASS (klass);
+	go_class->finalize = finalize;
+
+	style_class = GTK_STYLE_CLASS (klass);
 	style_class->draw_hline = draw_hline;
 	style_class->draw_vline = draw_vline;
 	style_class->draw_shadow = draw_shadow;
@@ -915,10 +851,12 @@ class_init (GceStyleClass *klass)
 	style_class->draw_focus = draw_focus;
 	style_class->draw_slider = draw_slider;
 	style_class->draw_handle = draw_handle;
+	/* FIXME: do own expander sometimes.
 	style_class->draw_expander = draw_expander;
-/*
-draw_layout
-*/
+	*/
+	/*
+	draw_layout
+	*/
 	style_class->draw_resize_grip = draw_resize_grip;
 }
 
